#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "memutil.h"
FILE	*safe_fopen( char *, char * );

static int c__3 = 3;

static int symnum(int natoms, REAL_T * amass, REAL_T * sn, int *linear)
{

/*     ----- routine to give the symmetry number. only for linear */
/*           molecules. for others symmetry number is unity ----- */


	/* Parameter adjustments */
	--amass;

	/* Function Body */
	*sn = 1.;
	if (natoms <= 2) {
		*linear = 1;
		if (amass[1] == amass[2]) {
			*sn = 2.;
		}
	}
	return 0;
}								/* symnum */

/* ----------------------------------------------------------------------- */

static int mofi(int natoms, REAL_T * c, REAL_T * amass, REAL_T * pmom)
{
	/* Initialized data */

	static char uplo[1] = "U";
	static char jobz[1] = "V";
	static REAL_T zero = 0.;

	/* System generated locals */

	/* Local variables */
	static REAL_T t[9], x, y, z, e2[30], wt;
	static int i, iat;
	static REAL_T com[3];
	static int ier, iaind;
	extern int dspev_(char *, char *, int *, REAL_T *,
					  REAL_T *, REAL_T *, int *, REAL_T *, int *);
	static REAL_T totwt, eigvec[9];


/*     compute the principal moments of inertia. */
/*     units are amu-bohr**2 */

	/* Parameter adjustments */
	--pmom;
	--amass;
	--c;

	/* Function Body */


/*     compute the position of the center of mass and translate */
/*     it to the origin. */

	com[0] = zero;
	com[1] = zero;
	com[2] = zero;

	totwt = zero;
	for (iat = 1; iat <= natoms; ++iat) {
		iaind = (iat - 1) * 3;
		wt = amass[iat];
		totwt += wt;
		com[0] += wt * c[iaind + 1];
		com[1] += wt * c[iaind + 2];
		com[2] += wt * c[iaind + 3];
	}

	com[0] /= totwt;
	com[1] /= totwt;
	com[2] /= totwt;

/*     compute the principal moments. */

	for (i = 1; i <= 9; ++i) {
		t[i - 1] = zero;
	}

	for (iat = 1; iat <= natoms; ++iat) {
		wt = amass[iat];
		x = c[1 + 3 * (iat - 1)] - com[1 - 1];
		y = c[2 + 3 * (iat - 1)] - com[2 - 1];
		z = c[3 + 3 * (iat - 1)] - com[3 - 1];
		t[0] += wt * (y * y + z * z);
		t[2] += wt * (x * x + z * z);
		t[5] += wt * (x * x + y * y);
		t[1] -= wt * x * y;
		t[3] -= wt * x * z;
		t[4] -= wt * y * z;
	}
	ier = 0;
	dspev_(jobz, uplo, &c__3, t, &pmom[1], eigvec, &c__3, e2, &ier);
	return 0;
}								/* mofi */

/* ----------------------------------------------------------------------- */

static int thermo(int natoms, int ilevel, REAL_T * c, REAL_T * amass,
		   REAL_T * freq, REAL_T * vtemp, REAL_T * evibn, REAL_T * cvibn,
		   REAL_T * svibn, REAL_T t, REAL_T patm, REAL_T energy )
{
	/* Initialized data */

	static REAL_T zero = 0.;
	static REAL_T akilo = 1e3;
	static REAL_T thresh = 900.;
	static int iout = 6;
	static REAL_T pstd = 101325.;
	static REAL_T pt2 = .2;
	static REAL_T half = .5;
	static REAL_T one = 1.;
	static REAL_T onept5 = 1.5;
	static REAL_T two = 2.;
	static REAL_T twopt5 = 2.5;
	static REAL_T four = 4.;
	static REAL_T eight = 8.;

	/* System generated locals */
	REAL_T d__1;

	/* Local variables */
	REAL_T e;
	int i;
	REAL_T p, s, pi, cv, sn, rt, em1;
	int iff;
	REAL_T arg, gas;
	int iat;
	REAL_T con, dum, dum1, dum2, argd, cvib, evib;
	int ndof;
	REAL_T avog, ezkc, pipi, ezpe, tokg, svib, crot, pmom[10],
		erot, etot, ctot, srot, stot, tovt, jpcal, tocal, ccont, ctran,
		econt, etran, scont, stran, tomet, rtemp, boltz, etovt, rtemp1,
		rtemp2, rtemp3, planck;
	int linear;
	REAL_T tokcal, hartre, weight;
	int nimag;


/*     given the structure of a molecule and its normal mode vibrational */
/*     frequencies this routine uses standard statistical mechanical */
/*     formulas for an ideal gas (in the canonical ensemble, see, */
/*     for example, d. a. mcquarrie, "statistical thermodynamics", */
/*     harper & row, new york, 1973, chapters 5, 6, and 8) to compute */
/*     the entropy, heat capacity, and internal energy. */

/*     the si system of units is used internally.  conversion to units */
/*     more familiar to most chemists is made for output. */


/*     amass:   atomic weights, in amu. */
/*     pmom:    principal moments of inertia, in amu-bohr**2 and */
/*              in ascending order. */
/*     freq:    vibrational frequencies, in cm**-1 and in ascending */
/*              order */
/*     c    :   coordinates in Angstroms */
/*     vtemp:   vibrational temperatures, in kelvin. */
/*     evibn:   contribution to e from the vibration n. */
/*     cvibn:   contribution to cv from the vibration n. */
/*     svibn:   contribution to s from the vibration n. */
/*     t:       temperature */
/*     patm:    pressure, in atmospheres */


	/* Parameter adjustments */
	--freq;
	--amass;
	--c;


/*     tokg:    kilograms per amu. */
/*     boltz:   boltzman constant, in joules per kelvin. */
/*     planck:  planck constant, in joule-seconds. */
/*     avog:    avogadro constant, in mol**(-1). */
/*     jpcal:   joules per calorie. */
/*     tomet:   metres per Angstrom. */
/*     hartre:  joules per hartree. */

	tokg = 1.660531e-27;
	boltz = 1.380622e-23;
	planck = 6.626196e-34;
	avog = 6.022169e23;
	jpcal = 4.18674;
	tomet = 1e-10;
	hartre = 4.35981e-18;

/*     compute the gas constant, pi, pi**2, and e. */
/*     compute the conversion factors cal per joule and kcal per joule. */

	gas = avog * boltz;
	pi = four * atan(one);
	pipi = pi * pi;
	e = exp(one);
	tocal = one / jpcal;
	tokcal = tocal / akilo;

/*     print the temperature and pressure. */

	p = pstd * patm;
	fprintf(stdout, "\n                - Thermochemistry -\n\n");
	fprintf(stdout, "Temperature:  %8.3f\n   Pressure:  %8.3f\n", t, patm);
	rt = gas * t;

/*     compute and print the molecular mass in amu, then convert to */
/*     kilograms. */

	weight = zero;
	for (iat = 1; iat <= natoms; ++iat) {
		weight += amass[iat];
	}
	fprintf(stdout, "       Mass:  %8.3f\n", weight);
	weight *= tokg;

/*     compute contributions due to translation: */
/*        etran-- internal energy */
/*        ctran-- constant v heat capacity */
/*        stran-- entropy */

	dum1 = boltz * t;
	d__1 = two * pi;
	dum2 = pow(d__1, onept5);
	arg = pow(dum1, onept5) / planck;
	arg = arg / p * (dum1 / planck);
	arg = arg * dum2 * (weight / planck);
	arg = arg * sqrt(weight) * pow(e, twopt5);
	stran = gas * log(arg);
	etran = onept5 * rt;
	ctran = onept5 * gas;

/*     Compute contributions due to electronic motion: */
/*        It is assumed that the first electronic excitation energy */
/*        is much greater than kt and that the ground state has a */
/*        degeneracy of one.  Under these conditions the electronic */
/*        partition function can be considered to be unity.  The */
/*        ground electronic state is taken to be the zero of */
/*        electronic energy. */

/*     for monatomics print and return. */

	if (natoms <= 1) {
		s = stran * tocal;
		e = etran * tokcal;
		cv = ctran * tocal;
		/* need print statement here */
		return 0;
	}

/*     compute contributions due to rotation. */

/*     Compute the principal moments of inertia, get the rotational */
/*     symmetry number, see if the molecule is linear, and compute */
/*     the rotational temperatures.  Note the imbedded conversion */
/*     of the moments to SI units. */

	mofi(natoms, &c[1], &amass[1], pmom);
	fprintf(stdout, "Principal moments of inertia in amu-A**2:\n");
	fprintf(stdout, "     %12.2f%12.2f%12.2f\n", pmom[0], pmom[1],
			pmom[2]);
	linear = 0;
	symnum(natoms, &amass[1], &sn, &linear);
	fprintf(stdout, "Rotational symmetry number is %2.0f\n", sn);
	con = planck / (boltz * eight * pipi);
	con = con / tokg * (planck / (tomet * tomet));
	if (linear) {
		rtemp = con / pmom[2];
		if (rtemp < pt2)
			fprintf(stdout, "Assuming classical behavior for rotation\n");
		fprintf(stdout, "Rotational temperature: %8.3f\n", rtemp);
	} else {
		rtemp1 = con / pmom[0];
		rtemp2 = con / pmom[1];
		rtemp3 = con / pmom[2];
		if (rtemp1 < pt2)
			fprintf(stdout, "Assuming classical behavior for rotation\n");
		fprintf(stdout, "Rotational temperatures: %8.3f  %8.3f  %8.3f\n",
				rtemp1, rtemp2, rtemp3);
	}

/*         erot-- rotational contribution to internal energy. */
/*         crot-- rotational contribution to cv. */
/*         srot-- rotational contribution to entropy. */

	if (linear) {
		erot = rt;
		crot = gas;
		arg = t / rtemp * (e / sn);
		srot = gas * log(arg);
	} else {
		erot = onept5 * rt;
		crot = onept5 * gas;
		arg = sqrt(pi * e * e * e) / sn;
		dum = t / rtemp1 * (t / rtemp2) * (t / rtemp3);
		arg *= sqrt(dum);
		srot = gas * log(arg);
	}

/*     compute contributions due to vibration. */

/*     compute vibrational temperatures and zero point vibrational */
/*     energy.  only real frequencies are included in the analysis. */

	nimag = 0;
	for( i=1; i<=3*natoms; i++ ) if( freq[i] < -0.5 ) nimag++;
	if( nimag > 0 ) fprintf( stdout, "found %d imaginary frequencies\n", nimag);

/*       (---iff is the first frequency to include in thermo:) */

	if (ilevel != 0) {
		iff = 0 + nimag;
	} else if (linear) {
		iff = 5 + nimag;
	} else {
		iff = 6 + nimag;
	}
	ndof = 3 * natoms - iff;
	con = planck / boltz;
	ezpe = zero;
	for (i = 1; i <= ndof; ++i) {
		vtemp[i] = freq[i + iff] * con * 3e10;
		ezpe += freq[i + iff] * 3e10;
	}
	ezpe = half * planck * ezpe;
	ezkc = ezpe * tokcal * avog;
	fprintf(stdout, "Zero-point vibrational energy: %10.3f\n", ezkc);


/*     compute: */
/*        evib-- the vibrational component of the internal energy. */
/*        cvib-- the vibrational component of the heat capacity. */
/*        svib-- the vibrational component of the entropy. */

	evib = zero;
	cvib = zero;
	svib = zero;
	for (i = 1; i <= ndof; ++i) {

/*       compute some common factors. */

		tovt = vtemp[i] / t;
		etovt = exp(tovt);
		em1 = etovt - one;

/*       compute contributions due to the i'th vibration. */

		econt = tovt * (half + one / em1);
		d__1 = tovt / em1;
		ccont = etovt * (d__1 * d__1);
		argd = one - one / etovt;
		if (argd > 1e-7) {
			scont = tovt / em1 - log(argd);
		} else {
			scont = 0.f;
			fprintf(stdout,
					"Warning: setting vibrational entropy to zero for mode %d with vtemp = %8.3f\n",
					i, vtemp[i]);
		}
		evibn[i] = econt * rt;
		cvibn[i] = ccont * gas;
		svibn[i] = scont * gas;

		evib += econt;
		cvib += ccont;
		svib += scont;
	}
	evib *= rt;
	cvib *= gas;
	svib *= gas;

/*     the units are now: */
/*         e-- joules/mol */
/*         c-- joules/mol-kelvin */
/*         s-- joules/mol-kelvin */

	etot = etran + erot + evib;
	ctot = ctran + crot + cvib;
	stot = stran + srot + svib;


/*     convert to the following and print */
/*         e-- kcal/mol */
/*         c-- cal/mol-kelvin */
/*         s-- cal/mol-kelvin */

	etran *= tokcal;
	ctran *= tocal;
	stran *= tocal;
	erot *= tokcal;
	crot *= tocal;
	srot *= tocal;
	evib *= tokcal;
	cvib *= tocal;
	svib *= tocal;
	etot = etran + erot + evib + energy;
	ctot = ctran + crot + cvib;
	stot = stran + srot + svib;
	for (i = 1; i <= ndof; ++i) {
		evibn[i] *= tokcal;
		cvibn[i] *= tocal;
		svibn[i] *= tocal;
	}

	fprintf(stdout,
			"\n             freq.         E               Cv              S\n");
	fprintf(stdout,
			"            cm**-1       kcal/mol       cal/mol-K     cal/mol-K\n");
	fprintf(stdout, "Total:               %11.3f    %11.3f    %11.3f\n",
			etot, ctot, stot);
	fprintf(stdout, "translational:       %11.3f    %11.3f    %11.3f\n",
			etran, ctran, stran);
	fprintf(stdout, "rotational:          %11.3f    %11.3f    %11.3f\n",
			erot, crot, srot);
	fprintf(stdout, "vibrational:         %11.3f    %11.3f    %11.3f\n",
			evib, cvib, svib);
	fprintf(stdout, "ff   energy:         %11.3f\n", energy );

	for (i = 1; i <= iff; ++i) {
		fprintf(stdout, " %5d   %10.3f\n", i, freq[i]);
	}
	for (i = 1; i <= ndof; ++i) {
		fprintf(stdout, " %5d   %10.3f  %11.3f    %11.3f    %11.3f\n",
				i + iff, freq[i + iff], evibn[i], cvibn[i], svibn[i]);
	}

	return 0;
}								/* thermo */

/* ----------------------------------------------------------------------- */

/* 
 * Compute normal modes
 *
 * Calling parameters are as follows:
 *
 * x[n]    contains the coordinates
 * n       dimension of variables
 * func    pointer to function that computes the energy, gradient and hessian
 *           (probably only works now with mme2())
 * eigp    number of modes to post to file "vecs"
 *
 */

int nmode( REAL_T *x, int *n, 
	    REAL_T ( *func)( REAL_T*, REAL_T*, REAL_T*, REAL_T*, int*, int* ),
		int *eigp )

{

  REAL_T *g, *m, *v, *h, *work;
  REAL_T sumg, temp, dgrad, energy;
  int niter, i, j, k, ncopy, natom, nrot, ret_val, odd, lwork, info;
  FILE *vfp;
  REAL_T pressure, temperature;

  char uplo, jobz;

  /* Allocate some dynamic vectors and matrices. */

  ncopy = *n;
  g = vector( 1, ncopy );
  m = vector( 1, ncopy );
  v = vector( 1, ncopy );
  h = vector( 0, ncopy*ncopy );
  work = vector( 0, 3*ncopy );
  niter = 1;

    /*
     * Compute the function value in "f",
     * its gradient (with respect to x) in "g", and
     * its hessian (with respect to x) in "h".
     * The atomic masses are returned in "m".
     * The number of atoms is returnd in "natom".
     *
     * The &x[1], &g[1] and &m[1] calling parameters
     * map from 1..n indexing in this nmode function
     * to 0..n-1 indexing in *func (which defaults to mme2).
     */

    energy = ( *func )( x, &g[ 1 ], h, &m[ 1 ], &natom, &niter );

    sumg = 0.0;
    for (i = 1; i <= ncopy; i++) {
      sumg += g[i] * g[i];
    }
	sumg = sqrt(sumg/ncopy);
	printf( "RMS gradient = %12.5e\n", sumg );
	fflush( stdout );

    /*
     * Mass weight the Hessian:
     */

	j = 1;
	for(i=1; i<=natom; i++) {
		g[j+2] = g[j+1] = g[j] = 1.0/sqrt(m[i]);
		j += 3;
	}

	k = 0;
	for(i=1; i<=ncopy; i++){
		for(j=1; j<=ncopy; j++){
			h[k] = g[i]*h[k]*g[j];
			k++;
		}
	}

    /*  Call the LAPACK routines; requires that LAPACK and BLAS support
        be available in your NAB installation    */
	if( *eigp ) jobz = 'V'; else jobz = 'N';
    uplo = 'U';
    lwork = 3*ncopy;
    k=0;

    dsyev_( &jobz, &uplo, n, h, n, &v[1], work, &lwork, &info );

	/* un-mass-weight the eigenvectors:  */
	k = 0;
	for( i=1; i<=ncopy; i++ ){
		for( j=1; j<=ncopy; j++ ){
			h[k++] *= g[j];
		}
	}

	/*  get the eigenvalues into cm**-1 units:   */
	for(i=1; i<=ncopy; i++){
		v[i] = v[i]>=0.0 ? 108.587*sqrt(v[i]) : -108.587*sqrt(-v[i]);
	}

	/* make an Amber-compatible formatted vecs file:   */
	if( *eigp ){
		vfp = safe_fopen( "vecs", "w" );
		fprintf( vfp, " modes from NAB\n" );
		fprintf( vfp, "%5d\n", ncopy );
		for( i=0; i<ncopy; i++ ){
			fprintf( vfp, "%11.5f", x[i] );
			if( i%7 == 6 ) fprintf( vfp, "\n" );
		}
		if( ncopy%7 != 0 ) fprintf( vfp, "\n" );
		k = 0;
		for( j=1; j<=*eigp; j++ ){
			fprintf( vfp, "****\n%5d%12.5f\n", j, v[j] );
			for( i=0; i<ncopy; i++ ){
				fprintf( vfp, "%11.5f", h[k+i] );
				if( i%7 == 6 ) fprintf( vfp, "\n" );
			}
			if( ncopy%7 != 0 ) fprintf( vfp, "\n" );
			k += ncopy;
		}
		fclose( vfp );
	}

	/*  get the thermodynamic parameters  */
	j = 0;
	pressure = 1.0;
	temperature = 298.15;
	thermo( natom, j, x, &m[1], &v[1], h, h+ncopy, h+2*ncopy, h+3*ncopy,
		temperature, pressure, energy);


  /* Free the dynamic vectors and matrices. */

  free_vector( g, 1, ncopy );
  free_vector( m, 1, ncopy );
  free_vector( v, 1, ncopy );
  free_vector( h, 0, ncopy*ncopy );
  free_vector( work, 0, 3*ncopy );

  ret_val = 0;
  return ret_val;
} 
